
#include "config.h"

#define INI_MAX_LINE 4096
#define INI_STOP_ON_FIRST_ERROR 1

#include "ini.h"

#define DECLARE_CONFIGBUF
#include "annex.h"

#include <stdlib.h>
#include <string.h>
#include <errno.h>
#include <fcntl.h>
#include <unistd.h>
#include <stdint.h>
#include <getopt.h>
#include <limits.h>

// parse an unsigned 64-bit number 
// uint64_str must contain *only* the text of the number 
// return the number, and set *success to true if we succeeded
// return -EINVAL if any of hte arguments were NULL
// return 0, LLONG_MAX, or LLONG_MIN if we failed
static uint64_t jwm_menu_parse_uint64( char const* uint64_str, bool* success ) {
   
   if( uint64_str == NULL || success == NULL ) {
      return -EINVAL;
   }
   
   char* tmp = NULL;
   uint64_t uid = (uint64_t)strtoll( uint64_str, &tmp, 10 );
   
   if( tmp == uint64_str ) {
      *success = false;
   }
   else {
      *success = true;
   }
   
   return uid;
}


// join two paths, writing the result to dest if dest is not NULL.
// otherwise, allocate and return a buffer containing the joined paths.
// return NULL on OOM 
static char* jwm_menu_fullpath( char const* root, char const* path, char* dest ) {

   char delim = 0;
   int path_off = 0;
   int len = 0;
   
   if( root == NULL || path == NULL ) {
      return NULL;
   }
   
   len = strlen(path) + strlen(root) + 2;

   if( strlen(root) > 0 ) {
      size_t root_delim_off = strlen(root) - 1;
      if( root[root_delim_off] != '/' && path[0] != '/' ) {
         len++;
         delim = '/';
      }
      else if( root[root_delim_off] == '/' && path[0] == '/' ) {
         path_off = 1;
      }
   }

   if( dest == NULL ) {
      dest = (char*)calloc( sizeof(char) * (len), 1 );
      if( dest == NULL ) {
         return NULL;
      }
   }

   memset(dest, 0, len);

   strcpy( dest, root );
   if( delim != 0 ) {
      dest[strlen(dest)] = '/';
   }
   strcat( dest, path + path_off );

   return dest;
}

// ini parser callback 
// return 1 on parsed (WARNING: masks OOM)
// return 0 on not parsed
static int jwm_menu_config_ini_parser( void* userdata, char const* section, char const* name, char const* value ) {
   
   struct jwm_menu_config* conf = (struct jwm_menu_config*)userdata;
   bool success = false;   
   
   if( strcmp( section, JWM_MENU_CONFIG_NAME ) == 0 ) {
      
      if( strcmp( name, JWM_MENU_CONFIG_PATH ) == 0 ) {
         
         if( conf->config_path == NULL ) {
            // save this 
            conf->config_path = strdup_or_null( value );
         }
         return 1;
      }
            
      if( strcmp( name, JWM_MENU_CONFIG_APPLICATION_LAUNCHER ) == 0 ) {
         
         if( conf->application_launcher == NULL ) {
            // save this 
            conf->application_launcher = strdup_or_null( value );
         }
         return 1;
      }
      
      if( strcmp( name, JWM_MENU_CONFIG_TERMINAL_EMULATOR ) == 0 ) {
         
         if( conf->terminal_emulator == NULL ) {
            // save this 
            conf->terminal_emulator = strdup_or_null( value );
         }
         return 1;
      }
      
      if( strcmp( name, JWM_MENU_CONFIG_FILE_MANAGER ) == 0 ) {
         
         if( conf->file_manager == NULL ) {
            // save this 
            conf->file_manager = strdup_or_null( value );
         }
         return 1;
      }
      
      if( strcmp( name, JWM_MENU_CONFIG_WEB_BROWSER ) == 0 ) {
         
         if( conf->web_browser == NULL ) {
            // save this 
            conf->web_browser = strdup_or_null( value );
         }
         return 1;
      }
      
      if( strcmp( name, JWM_MENU_CONFIG_MAIL_CLIENT ) == 0 ) {
         
         if( conf->mail_client == NULL ) {
            // save this 
            conf->mail_client = strdup_or_null( value );
         }
         return 1;
      }
      
      if( strcmp( name, JWM_MENU_CONFIG_LOGOUT_DIALOG ) == 0 ) {
         
         if( conf->logout_dialog == NULL ) {
            // save this 
            conf->logout_dialog = strdup_or_null( value );
         }
         return 1;
      }

      fprintf(stderr, "Unrecognized field '%s'\n", name);      
      return 1;
   }
   
   return 1;
}

// config sanity check 
int jwm_menu_config_sanity_check( struct jwm_menu_config* conf ) {
   
   int rc = 0;
   
   if( conf->config_path == NULL ) {
      
      fprintf(stderr, "[ERROR]: missing config_path\n");
      rc = -EINVAL;
   }
   
   if( conf->application_launcher == NULL ) {
      
      fprintf(stderr, "[ERROR]: missing application_launcher\n");
      rc = -EINVAL;
   }
   
   if( conf->terminal_emulator == NULL ) {
      
      fprintf(stderr, "[ERROR]: missing terminal_emulator\n");
      rc = -EINVAL;
   }
   
   if( conf->file_manager == NULL ) {
      
      fprintf(stderr, "[ERROR]: missing file_manager\n");
      rc = -EINVAL;
   }
   
   if( conf->web_browser == NULL ) {
      
      fprintf(stderr, "[ERROR]: missing web_browser\n");
      rc = -EINVAL;
   }
   
   if( conf->mail_client == NULL ) {
      
      fprintf(stderr, "[ERROR]: missing mail_client\n");
      rc = -EINVAL;
   }
   
   if( conf->logout_dialog == NULL ) {
      
      fprintf(stderr, "[ERROR]: missing logout_dialog\n");
      rc = -EINVAL;
   }
   
   return rc;
}
   

// initialize a config 
// always succeeds
int jwm_menu_config_init( struct jwm_menu_config* conf )
{
   memset( conf, 0, sizeof(struct jwm_menu_config) );
   return 0;
}

// load from a file, by path
// return on on success
// return -errno on failure to open 
int jwm_menu_config_load( char const* path, struct jwm_menu_config* conf )
{
   
   FILE* f = NULL;
   int rc = 0;
   
   f = fopen(path, "r");
   if(!f) {
      rc = -errno;
      return rc;
   }
   
   rc = jwm_menu_config_load_file(f, conf);
   
   fclose(f);
   
   return rc;
}

// load from a file
// return 0 on success
// return -errno on failure to load
int jwm_menu_config_load_file( FILE* file, struct jwm_menu_config* conf ) {
   
   int rc = 0;
   
   rc = ini_parse_file( file, jwm_menu_config_ini_parser, conf );
   if( rc != 0 ) {
      fprintf(stderr, "ini_parse_file(config) rc = %d\n", rc );
      jwm_menu_config_free( conf );

      return rc;
   }

   // convert paths 
   rc = jwm_menu_config_fullpaths( conf );
   if( rc != 0 ) {

      fprintf(stderr, "jwm_menu_config_fullpaths: %s\n", strerror(-rc) );
      jwm_menu_config_free( conf );
      return rc;
   }
   
   return rc;
}

// free a config
// always succeeds
int jwm_menu_config_free( struct jwm_menu_config* conf )
{
   if( conf->config_path != NULL ) {
      
      free( conf->config_path );
      conf->config_path = NULL;
   }
   
   return 0;
}

// convert all paths in the config to absolute paths 
// return 0 on success 
// return -ENOMEM on OOM 
// return -ERANGE if cwd is too long
int jwm_menu_config_fullpaths( struct jwm_menu_config* conf ) {
   
   char** need_fullpath[] = {
      &conf->config_path,
      NULL
   };
   
   char cwd_buf[ PATH_MAX + 1 ];
   memset( cwd_buf, 0, PATH_MAX + 1 );
   
   char* tmp = getcwd( cwd_buf, PATH_MAX );
   if( tmp == NULL ) {
      
      fprintf(stderr, "Current working directory exceeds %u bytes\n", PATH_MAX);
      return -ERANGE;
   }
   
   for( int i = 0; need_fullpath[i] != NULL; i++ ) {
      
      if( need_fullpath[i] != NULL && (*need_fullpath[i]) != NULL ) {
         
         if( *(need_fullpath[i])[0] != '/' ) {
               
            // relative path 
            char* new_path = jwm_menu_fullpath( cwd_buf, *(need_fullpath)[i], NULL );
            if( new_path == NULL ) {
               
               return -ENOMEM;
            }
            
            free( *(need_fullpath[i]) );
            *(need_fullpath[i]) = new_path;
         }
      }
   }
   
   return 0;
}
