/*
 * rsvg-convert.c
 * 
 * Forked from https://github.com/GNOME/librsvg
 * 
 * All modifications to the original source file are:
 *  - Copyright (C) 2022 Aitor Cuadrado Zubizarreta <aitor_czr@gnuinos.org> 
 */

/* -*- Mode: C; indent-tabs-mode: nil; c-basic-offset: 4 -*- */
/* vim: set sw=4 sts=4 expandtab: */
/*
 
   Copyright (C) 2005 Red Hat, Inc.
   Copyright (C) 2005 Dom Lachowicz <cinamod@hotmail.com>
   Copyright (C) 2005 Caleb Moore <c.moore@student.unsw.edu.au>
   Copyright (C) 2019 Federico Mena Quintero <federico@gnome.org>
  
   This library is free software; you can redistribute it and/or
   modify it under the terms of the GNU Lesser General Public
   License as published by the Free Software Foundation; either
   version 2.1 of the License, or (at your option) any later version.

   This library is distributed in the hope that it will be useful,
   but WITHOUT ANY WARRANTY; without even the implied warranty of
   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
   Lesser General Public License for more details.

   You should have received a copy of the GNU Lesser General Public
   License along with this library; if not, write to the Free Software
   Foundation, Inc., 51 Franklin Street, Fifth Floor, Boston, MA  02110-1301  USA
  
   Authors: Carl Worth <cworth@cworth.org>, 
            Caleb Moore <c.moore@student.unsw.edu.au>,
            Dom Lachowicz <cinamod@hotmail.com>,
            Federico Mena Quintero <federico@gnome.org>
*/

#include <math.h>
#include <errno.h>
#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include <locale.h>
#include <glib/gi18n.h>
#include <gio/gio.h>
#include <gio/gunixinputstream.h>
#include <librsvg/rsvg.h>


typedef enum {
    SIZE_KIND_ZOOM,
    SIZE_KIND_WH,
    SIZE_KIND_WH_MAX,
    SIZE_KIND_ZOOM_MAX
} SizeKind;

typedef struct {
    SizeKind kind;
    double x_zoom;
    double y_zoom;
    gint width;
    gint height;

    gboolean keep_aspect_ratio;
} SizeMode;

static void
get_final_size (int *width, int *height, SizeMode *real_data)
{ 
    double zoomx, zoomy, zoom;

    int in_width, in_height;

    in_width = *width;
    in_height = *height;

    switch (real_data->kind) {
    case SIZE_KIND_ZOOM:
        if (*width < 0 || *height < 0)
            return;

        *width = floor (real_data->x_zoom * *width + 0.5);
        *height = floor (real_data->y_zoom * *height + 0.5);
        break;

    case SIZE_KIND_ZOOM_MAX:
        if (*width < 0 || *height < 0)
            return;

        *width = floor (real_data->x_zoom * *width + 0.5);
        *height = floor (real_data->y_zoom * *height + 0.5);

        if (*width > real_data->width || *height > real_data->height) {
            zoomx = (double) real_data->width / *width;
            zoomy = (double) real_data->height / *height;
            zoom = MIN (zoomx, zoomy);

            *width = floor (zoom * *width + 0.5);
            *height = floor (zoom * *height + 0.5);
        }
        break;

    case SIZE_KIND_WH_MAX:
        if (*width < 0 || *height < 0)
            return;

        zoomx = (double) real_data->width / *width;
        zoomy = (double) real_data->height / *height;
        if (zoomx < 0)
            zoom = zoomy;
        else if (zoomy < 0)
            zoom = zoomx;
        else
            zoom = MIN (zoomx, zoomy);

        *width = floor (zoom * *width + 0.5);
        *height = floor (zoom * *height + 0.5);
        break;

    case SIZE_KIND_WH:
        if (real_data->width != -1)
            *width = real_data->width;
        if (real_data->height != -1)
            *height = real_data->height;
        break;

    default:
        g_assert_not_reached ();
    }

    if (real_data->keep_aspect_ratio) {
        int out_min = MIN (*width, *height);

        if (out_min == *width) {
            *height = in_height * ((double) *width / (double) in_width);
        } else {
            *width = in_width * ((double) *height / (double) in_height);
        }
    }
}

static void
display_error (GError * err)
{
    if (err) {
        g_printerr ("%s\n", err->message);
        g_error_free (err);
    }
}

static cairo_status_t
rsvg_cairo_write_func (void *closure, const unsigned char *data, unsigned int length)
{
    if (fwrite (data, 1, length, (FILE *) closure) == length)
        return CAIRO_STATUS_SUCCESS;
    return CAIRO_STATUS_WRITE_ERROR;
}

static char *
get_lookup_id_from_command_line (const char *lookup_id)
{
    char *export_lookup_id;

    if (lookup_id == NULL)
        export_lookup_id = NULL;
    else {
        /* rsvg_handle_has_sub() expects ids to have a '#' prepended to them, so
         * it can lookup ids in externs like "subfile.svg#subid".  For the
         * user's convenience, we include this '#' automatically; we only
         * support specifying ids from the toplevel, and don't expect users to
         * lookup things in externs.
         */
        export_lookup_id = g_strdup_printf ("#%s", lookup_id);
    }

    return export_lookup_id;
}
 
void rsvg_convert(const char *input, const char *output)
{ 
    double dpi_x = 90.0;
    double dpi_y = 90.0;
    char *export_id = NULL;
    
    guint32 background_color = 0;
    GError *error = NULL;
	SizeMode size_data;
	
	GFile *file;
	GInputStream *stream;

    RsvgHandle *rsvg = NULL;
    cairo_surface_t *surface = NULL;
    cairo_t *cr = NULL;
    RsvgDimensionData dimensions;
    FILE *output_file = stdout;
    
    char *export_lookup_id;
    double unscaled_width, unscaled_height;
    int scaled_width, scaled_height;

    /* Set the locale so that UTF-8 filenames work */
    setlocale(LC_ALL, "");

    if (output != NULL) {
        output_file = fopen (output, "wb");
        if (!output_file) {
            g_printerr (_("Error saving to file: %s\n"), output);
            exit (EXIT_FAILURE);
        }
    }

	file = g_file_new_for_commandline_arg (input);
	stream = (GInputStream *) g_file_read (file, NULL, &error);
	if(!stream)
		goto done;

	rsvg = rsvg_handle_new_from_stream_sync (stream, file, RSVG_HANDLE_FLAGS_NONE, NULL, &error);

done:
	g_clear_object (&stream);
	g_clear_object (&file);

	if (error != NULL) {
		g_printerr (_("Error reading SVG:"));
		display_error (error);
		g_printerr ("\n");
		exit (EXIT_FAILURE);
	}

	g_assert (rsvg != NULL);

	rsvg_handle_set_dpi_x_y (rsvg, dpi_x, dpi_y);

	export_lookup_id = get_lookup_id_from_command_line (export_id);
	if (export_lookup_id != NULL && !rsvg_handle_has_sub (rsvg, export_lookup_id)) {
		g_printerr (_("File %s does not have an object with id \"%s\"\n"), input, export_id);
		exit (EXIT_FAILURE);
	}

	if (!rsvg_handle_get_dimensions_sub (rsvg, &dimensions, export_lookup_id)) {
		g_printerr ("Could not get dimensions for file %s\n", input);
		exit (EXIT_FAILURE);
	}

	if (dimensions.width == 0 || dimensions.height == 0) {
		g_printerr ("The SVG %s has no dimensions\n", input);
		exit (EXIT_FAILURE);
	}

	unscaled_width = dimensions.width;
	unscaled_height = dimensions.height;

	size_data.kind = SIZE_KIND_ZOOM;
	size_data.x_zoom = 1.0;
	size_data.y_zoom = 1.0;
	size_data.keep_aspect_ratio = 0;

	scaled_width = dimensions.width;
	scaled_height = dimensions.height;
	get_final_size (&scaled_width, &scaled_height, &size_data);

	surface = cairo_image_surface_create (CAIRO_FORMAT_ARGB32, scaled_width, scaled_height);

	cr = cairo_create (surface);
	cairo_scale (cr, scaled_width / unscaled_width, scaled_height / unscaled_height);

	if (!rsvg_handle_render_cairo_sub (rsvg, cr, export_lookup_id)) {
		g_printerr ("Could not render file %s\n", input);
		exit (1);
	}

	g_free (export_lookup_id);

	cairo_surface_write_to_png_stream (surface, rsvg_cairo_write_func, output_file);

	g_object_unref (rsvg);
    cairo_destroy (cr);
    cairo_surface_destroy (surface);
    fclose (output_file);
}
